/******************************************************************************
 *  Name:        Nayana Prasad Nagendra
 *  NetID:       nagendra
 *  Precept:     
 * 
 *  Description: Provides logging feature to LCN
 *               Provides a method to convert any LCN object into a string
 *               Provides testing feature
 *
 ******************************************************************************/

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.FileAlreadyExistsException;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Arrays;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import edu.princeton.cs.algs4.In;
import edu.princeton.cs.algs4.StopwatchCPU;
import edu.princeton.cs.algs4.StdRandom;

public class LCNTools {

    private static String file;
    private static LCN student;
    private static int[][] board;
    private static int ref_solution;
    private static boolean ref_isASolution;
    
    // Takes a LCN and returns a string representation of it
    public static String toString(LCN lcn) {
        StringBuilder s = new StringBuilder();
        s.append(lcn.n + "\n");
        for (int i = 0; i < lcn.n; i++) {
            for (int j = 0; j < lcn.n; j++) {
                s.append(String.format("%2d ", lcn.board[i][j]));
            }
            s.append("\n");
        }
        s.append("isASolution: " + lcn.isASolution());
        s.append(", Solution: " + lcn.solution() + "\n");
        return s.toString();
    }
    

    // private method to copy values from input file to an array
    private static int[][] readFromFile() {
        In in = new In(file);
        int sz = in.readInt();
        int[][] tiles = new int[sz][sz];
        for (int i = 0; i < sz; i++) {
            for (int j = 0; j < sz; j++) {
                tiles[i][j] = in.readInt();
            }
        }
        return tiles;
    }

    // create random n-by-n array of tiles
    private static int[][] createBoard(int n) {
        int[][] tiles = new int[n][n];
        for (int i = 0; i < n; i++)
            for (int j = 0; j < n; j++)
                tiles[i][j] = n*i + j;
        return tiles;
    }

    // tests the correctness of solution() method
    private static void testSolution() {

        System.out.println("Test 1: Testing correctness of solution()");

        int stu_solution = student.solution();
        
        // create a String to display the test to a human being
        String test_caption = stu_solution + " (obtained) != " + ref_solution + " (desired)";
        
        // check if the resulting value is same as reference value
        boolean test_success = (stu_solution == ref_solution);
        
        if (!test_success) {
            System.out.println("FAILED: solution() "
                               + " : " + test_caption);
        }
        else {
            System.out.println("passed: solution()");
        }
    }

    // tests the correctness of isASolution() method
    private static void testIsASolution() {

        System.out.println("Test 2: Testing correctness of isASolution()");

        boolean stu_isASol = student.isASolution();
        
        // create a String to display the test to a human being
        String test_caption = stu_isASol + " (obtained) != " + ref_isASolution + " (desired)";
        
        // check if the resulting value is same as reference value
        boolean test_success = (stu_isASol == ref_isASolution);
        
        if (!test_success) {
            System.out.println("FAILED: isASolution() "
                               + " : " + test_caption);
        }
        else {
            System.out.println("passed: isASolution()");
        }
    }

    // time constructor in LCN
    private static void timeConstructor() {
        try {
            int MAX_N = 1024;
            double MAX_SECONDS = 5.0;

            System.out.println("-----------------------------------------");
            System.out.println("Time constructor on random n-by-n arrays.");
            System.out.println("-----------------------------------------");

            double old = 0.0, time = 0.0;
            System.out.printf("%7s %5s %8s %9s\n", "", "N", "seconds", "log ratio");
            System.out.printf("%7s %s\n", "", "------------------------");

            for (int i = 2; true; i++) {
                int n = (int) Math.round(Math.pow(Math.sqrt(2.0), i));
                old = time;
                int[][] tiles = createBoard(n);

                StopwatchCPU stopwatch = new StopwatchCPU();      
                LCN student = new LCN(tiles);
                time = stopwatch.elapsedTime();
                
                double ratio = 2 * Math.log(time / old) / Math.log(2);
                System.out.printf("%7s %5d %8.3f %9.1f\n", "", n, time, ratio);

                if (n >= MAX_N) {
                    System.out.println("passed: Constructor timing");
                    return;
                }

                if (old + time > MAX_SECONDS)
                    break;
            }

            System.out.println("FAILED: Constructor timing");
        }
        catch (Throwable e) {
            System.out.println(e);
        }
    }

    private static double run(int N) {
        try {
            int[][] tiles = createBoard(N);
            LCN stu = new LCN(tiles);
            StopwatchCPU stopwatch = new StopwatchCPU();
            for (int i = 0; i < N; i++) {
                stu.solution();
            }

            for (int i = 0; i < N; i++) {
                double r = StdRandom.uniform();
                if (r < 0.50) stu.isASolution();
                else stu.solution();
            }

            return stopwatch.elapsedTime();
        }
        catch (Throwable e) {
            System.out.println(e);
        }
        return Double.POSITIVE_INFINITY;
    }

    private static void timeSolution() {
        int[] testNs = { 512, 1024, 2048, 4096 };
        double MAX_SECONDS = 5.0;

        System.out.println("-----------------------------------------------------------------");
        System.out.println("Time solution() and isASolution() methods on random n-by-n arrays");
        System.out.println("-----------------------------------------------------------------");

        System.out.printf("%21s %8s\n", "N", "seconds");
        System.out.printf("%15s %s\n", "", "-------------------");
        for (int i = 0; i < testNs.length; i++) {
            int N = testNs[i];

            // do it twice to get more reliable answers
            double time1 = run(N);
            double time2 = run(N);

            if (time1 + time2 > MAX_SECONDS) {
                System.out.printf("=> FAILED    %8d %8.2f\n   [ Most likely one of your operations is not constant time. ]\n", N, time2);
                return;
            }

            //anything less than maxTime seconds is a pass
            System.out.printf("=> passed    %8d %8.4f\n", N, time2);
        }

        System.out.println("passed: solution() and isASolution() timing");
    }

    public static void test(String filename, int ref_sol, boolean ref_isASol) {
        file = filename;
        board = readFromFile();
        student = new LCN(board);
        ref_solution = ref_sol;
        ref_isASolution = ref_isASol;

        System.out.println("----------------------------------------------------------");
        System.out.println("Input file: " + filename + " along with your method return values:");
        System.out.println("----------------------------------------------------------");
        System.out.println(toString(student));
        System.out.println();

        System.out.println("Testing the file " + filename);
        System.out.println("----------------------------");
        System.out.println();

        testSolution();
        System.out.println();

        testIsASolution();
        System.out.println();

    }

    public static void test() {
        test("demo.txt", 5, true);
        test("puzzle50.txt", 2, true);
        test("fail.txt", Integer.MIN_VALUE, false);

        timeConstructor();
        System.out.println();

        timeSolution();
        System.out.println();

    }

    // To log the current java file.
    // This is done by first creating a directory with timestamp as name
    // within the "logs" directory and copying the current version of 
    // LCN.java inside it.
    public static void logFile() {
        
        // Check if there are any changes in LCN.java
        // compared to the previous log.
        // All logs are saved in timestamped sub-directory in the
        // current directory, with the format MM_dd_HH_MM_ss_SSS
        boolean createNewLog = true;
        boolean createLogDir = true;

        File pwd = new File(".");
        File[] listOfFiles = pwd.listFiles();
        for (File subFile : listOfFiles) {
            String name = subFile.getName();
            if (subFile.isDirectory() && subFile.getName().equals("logs")) {
                createLogDir = false;
                break;
            }
        }
        
        if (!createLogDir) {
            File logsDir = new File("logs/");
            File[] directories = logsDir.listFiles();
            if (directories != null) {
                try {
                    for (int dir = 0; dir < directories.length; dir++) {
                        Path origPath = Paths.get("LCN.java");
                        Path prevLogPath = Paths.get(directories[dir]+"/LCN.java");

                        byte[] orig = Files.readAllBytes(origPath);
                        byte[] prevLog = Files.readAllBytes(prevLogPath);
                        if (Arrays.equals(orig, prevLog)) {
                           createNewLog = false;
                           break;
                        }
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
       
        // Save the log only if there are changes compared to
        // the immediate previous log
        if (createNewLog || createLogDir) {

            Path logPath = Paths.get("logs");

            // get current timestamp
            String time_pattern = "MM_dd_HH_mm_ss_SSS";
            SimpleDateFormat sdf = new SimpleDateFormat(time_pattern);
            String now = sdf.format(new Date());

            // create a subdirectory with the timestamp as the name
            Path path = Paths.get("logs/"+now);       

            try {
                if (createLogDir) {    
                    // First time run, create the "logs" directory
                    Path logDir = Files.createDirectory(logPath);
                }
                Path newDir = Files.createDirectory(path);

                // copy LCN.{java,class} into the timestamped directory
                Path sourceJavaPath = Paths.get("LCN.java");
                Path destJavaPath = Paths.get("logs/"+now+"/LCN.java") ;
                Path sourceClassPath = Paths.get("LCN.class");
                Path destClassPath = Paths.get("logs/"+now+"/LCN.class") ;

                Files.copy(sourceJavaPath, destJavaPath);
                Files.copy(sourceClassPath, destClassPath);

                // Now zip the entire "logs" directory as "logs.zip"
                // Let us first delete the previous version of zip file,
                // if it exists
                File logsZip = new File("logs.zip");
                if (logsZip.exists()) {
                    assert(logsZip.delete());
                }

                Path logsPath = Paths.get("logs.zip");
                OutputStream fos = Files.newOutputStream(logsPath);
                ZipOutputStream zos = new ZipOutputStream(fos);

                // Recursively zip the files
                addToZip(zos, new File("logs/"), null);
                zos.flush();
                fos.flush();
                zos.close();
                fos.close();

            } catch (IOException e) {
                //something else went wrong
                e.printStackTrace();
            }
        }

    }

    // Recursively zip the files inside a directory
    public static void addToZip(ZipOutputStream zos, File file, String parent) 
    throws IOException{

        if (file == null || !file.exists()) {
            return;
        }

        String zipName = file.getName();
        if (parent!=null && !parent.isEmpty()) {
            zipName = parent + "/" + file.getName();
        }

        if (file.isDirectory()) {
            File[] listOfFiles = file.listFiles();
            if (listOfFiles != null) {
                for (File subFile : listOfFiles) {
                    addToZip(zos, subFile, zipName);
                }
            }
        } else {
            byte[] buffer = new byte[1024];
            InputStream fis = Files.newInputStream(file.toPath());
            zos.putNextEntry(new ZipEntry(zipName));
            int length;
    
            while ((length = fis.read(buffer)) > 0) {
                zos.write(buffer, 0, length);
            }

            zos.closeEntry();
            fis.close();
        } 
    }
}
